import com.mindfusion.diagramming.*;
import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;
import java.io.File;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Parameter;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;


/**
 * Created by MindFusion on 12/19/2016.
 */
public class MainWindow extends JFrame implements ActionListener {

    JButton openButton;
    JFileChooser fileChooser;
    String jarFileName;
    File currFile;

    private Diagram diagram;
    private DiagramView diagramView;
    private JScrollPane _scrollPane;
    private ZoomControl zoomer;

    double coef = 3.5;

    public MainWindow() {
        super("Class Diagram Sample");
        Image diagramImage = createImageFromPath("res/diagram.png");
        this.setIconImage(diagramImage);
        setVisible(true);
        setSize(300, 300);

        currFile = new File(System.getProperty("user.dir"));

        openButton = new JButton("Load *.jar file");
        //Use the default text position of CENTER, TRAILING (RIGHT).
        openButton.setMnemonic(KeyEvent.VK_E);

        //Listen for actions on buttons 1 and 3.
        openButton.addActionListener(this);
        openButton.setToolTipText("Choose a *.jar file.");

        //diagram initialization
        diagram = new Diagram();
        diagram.setAutoResize(AutoResize.RightAndDown);

        //initialize a diagramView that will render the diagram.
        diagramView = new DiagramView(diagram);
        diagramView.setVisible(true);

        //provide a zoomer for the diagram
        zoomer = new ZoomControl();
        zoomer.setView(diagramView);
        zoomer.setPreferredSize(new Dimension(70, 50));
        zoomer.setVisible(true);

        //use a scroll pane to host large diagrams
        _scrollPane = new JScrollPane(diagramView);
        _scrollPane.setVisible(true);
        _scrollPane.setAutoscrolls(true);

        //set a BorderLayout
        getContentPane().setLayout(new BorderLayout());
        this.add(zoomer, BorderLayout.EAST);
        this.add(createLegendPanel(), BorderLayout.SOUTH);

        //add the JScrollPane last and assign to it all available space
        this.add(_scrollPane, BorderLayout.CENTER);
        this.setJMenuBar(createMenuBar());

        //the JFrame starts maximized
        setExtendedState(java.awt.Frame.MAXIMIZED_BOTH);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    }

    //creates the panel with the diagram legend
    private JPanel createLegendPanel()
    {
        //the panel uses BoxLayout
        JPanel controlsPanel = new JPanel();
        controlsPanel.setLayout(new BoxLayout(controlsPanel, BoxLayout.LINE_AXIS));
        controlsPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
        controlsPanel.add(Box.createHorizontalGlue());

        //initialize the images for the legend
        Image constructorImage = createImageFromPath("res/constructor.png");
        Image methodImage = createImageFromPath("res/method.png");
        Image fieldImage = createImageFromPath("res/field.png");
        Image blueImage = createImageFromPath("res/blue.png");
        Image greenImage = createImageFromPath("res/green.png");
        Image yellowImage = createImageFromPath("res/yellow.png");

        //create icons from the images
        ImageIcon constructorIcon = new ImageIcon(constructorImage, "Constructor");
        ImageIcon methodIcon = new ImageIcon(methodImage, "Method");
        ImageIcon fieldIcon = new ImageIcon(fieldImage, "Field");
        ImageIcon blueIcon = new ImageIcon(blueImage, "Class");
        ImageIcon greenIcon = new ImageIcon(greenImage, "Enum");
        ImageIcon yellowIcon = new ImageIcon(yellowImage, "Interface");

        //create labels for each item on the legend
        JLabel label = new JLabel("Constructor", constructorIcon, JLabel.CENTER);
        label.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        controlsPanel.add(label);
        controlsPanel.add(Box.createRigidArea(new Dimension(10, 0)));
        controlsPanel.add(Box.createHorizontalGlue());

        label = new JLabel("Method", methodIcon, JLabel.CENTER);
        label.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        controlsPanel.add(label);
        controlsPanel.add(Box.createRigidArea(new Dimension(10, 0)));
        controlsPanel.add(Box.createHorizontalGlue());

        label = new JLabel("Field", fieldIcon, JLabel.CENTER);
        label.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        controlsPanel.add(label);
        controlsPanel.add(Box.createRigidArea(new Dimension(50, 0)));
        controlsPanel.add(Box.createHorizontalGlue());

        label = new JLabel("Class", blueIcon, JLabel.CENTER);
        label.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        controlsPanel.add(label);
        controlsPanel.add(Box.createRigidArea(new Dimension(10, 0)));
        controlsPanel.add(Box.createHorizontalGlue());

        label = new JLabel("Interface", yellowIcon, JLabel.CENTER);
        label.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        controlsPanel.add(label);
        controlsPanel.add(Box.createRigidArea(new Dimension(10, 0)));
        controlsPanel.add(Box.createHorizontalGlue());

        label = new JLabel("Enum", greenIcon, JLabel.CENTER);
        label.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        controlsPanel.add(label);
        controlsPanel.add(Box.createRigidArea(new Dimension(10, 0)));
        controlsPanel.add(Box.createHorizontalGlue());


        return controlsPanel;


    }

    /**
     * Creates the menu bar for the application.
     *
     * @return      the JMenuBar that was created
     */
    private JMenuBar createMenuBar() {
        JMenuBar menuBar;
        JMenu menu;
        JMenuItem menuItem;

        //Create the menu bar.
        menuBar = new JMenuBar();

        //Build the first menu.
        menu = new JMenu("File");
        menu.setMnemonic(KeyEvent.VK_F);
        menu.getAccessibleContext().setAccessibleDescription(
                "File operations");
        menuBar.add(menu);

        //initialize the 'Open' submenu
        menuItem = new JMenuItem("Open JAR", KeyEvent.VK_O);
        menuItem.setAccelerator(KeyStroke.getKeyStroke(
                KeyEvent.VK_O, ActionEvent.ALT_MASK));
        //assign ActionListener that will handle events
        menuItem.addActionListener(this);
        //a command identifier
        menuItem.setActionCommand("open_jar");
        menu.add(menuItem);

        return menuBar;
    }

    //the method that handles events
    public void actionPerformed(ActionEvent e)
    {
        //identify the command
        if ("open_jar".equals(e.getActionCommand())) {

            //a list with MemberInfo objects that hold class info
            ArrayList<MemberInfo> membersList = new ArrayList<MemberInfo>();
            fileChooser = new JFileChooser();

            //set the default directory to this file's directory
            fileChooser.setCurrentDirectory(currFile);

            //filter only *.jar files
            FileFilter filter = new FileNameExtensionFilter(null, "jar");
            fileChooser.setFileFilter(filter);
            fileChooser.removeChoosableFileFilter(fileChooser.getAcceptAllFileFilter());

            int response = fileChooser.showOpenDialog(null);
            //if the user has selected a *.jar file - pass it for parsing.
            if (response == JFileChooser.APPROVE_OPTION) {
                currFile = fileChooser.getSelectedFile();
                jarFileName = currFile.getPath();

                //check if this is a valid file name
                if (jarFileName.length() > 0)
                    membersList = readClasses(jarFileName);

                //create elements from the members of the *.jar file.
                if (membersList.size() > 0)
                    createDiagramElements(membersList);
            }
        }
    }

    /**
     * A helper method. Creates an image from a given path.
     * @param path   The url of the image location.
     * @return       The image that was initialized or null if the opration fails.
     */
    private Image createImageFromPath (String path) {
        File pathToFile = new File(path);
        Image img = null;


        try {
            img = ImageIO.read(pathToFile);

        } catch (IOException ex) {
            ex.printStackTrace();
        }

        return img;
    }


    /**
     * Creates diagram elements from the file info that was provided.
     *
     *
     * @param membersList  A list with MemberInfo objects that describe each file in the *.jar.
     */
    private void createDiagramElements (ArrayList<MemberInfo> membersList)
    {

        //initialize the legend images used for each class member.
        Image constructorImage = createImageFromPath("res/constructor.png");
        Image methodImage = createImageFromPath("res/method.png");
        Image fieldImage = createImageFromPath("res/field.png");

        //keep separate lists for the interfaces and enums.
        ArrayList<TableNode> interfaces = new ArrayList<TableNode>();
        ArrayList<TableNode> enums = new ArrayList<TableNode>();

        //specify that the connection between diagram items is between tables.
        diagram.setTableConnectionStyle(TableConnectionStyle.Table);

        for(MemberInfo info : membersList)
        {
            //create the TableNode
            TableNode _table = diagram.getFactory().createTableNode(10, 10, 5, 5, 2, info.memberCount());
            //set the name of the member and the package as table caption.
            String caption = "<b>" + info.getName() + "</b>";
            if(info.getPackageName().length() > 0)
                caption += "\n" + info.getPackageName();
            _table.setCaption(caption);
            //center the caption
            _table.setCaptionFormat((new TextFormat(Align.Center, Align.Center)));
            //increase the default caption height
            _table.setCaptionHeight(10f);

            //mo shadow for the table nodes.
            _table.setShadowOffsetX(0f);
            _table.setShadowOffsetY(0f);
            //specify an id
            _table.setId(info.getFullName());

            //add enums to the list and paint them green
            if(info.isEnum()) {
                _table.setBrush(new SolidBrush(new Color((int) 210, (int) 250, (int) 208)));
                enums.add(_table);
            }//add interfaces to the list and paint them yellow
            else if(info.isInterface()) {
                _table.setBrush(new SolidBrush(new Color((int) 250, (int) 235, (int) 140)));
                interfaces.add(_table);
            }
            else //the class nodes are blue
                _table.setBrush(new SolidBrush(new Color((int)197, (int)223, (int)238)));


            int index = 0;
            //fill the first cells with data for the constructors.
            for (String constructor : info.getConstructors())
            {
                _table.getCell(0, index).setImage(constructorImage);
                _table.getCell(1, index).setText(constructor);
                index++;
            }

            //table rows with data for the methods.
            for (String method : info.getMethods())
            {
                _table.getCell(0, index).setImage(methodImage);
                if (info.isInterface())
                	_table.getCell(1, index).getTextFormat().setNoWrap(true);
                _table.getCell(1, index).setText(method);
                index++;
            }

            //table rows with data for the fields.
           for (String field : info.getFields())
           {
                _table.getCell(0, index).setImage(fieldImage);
                _table.getCell(1, index).setText(field);
                index++;
            }

            //adjust each cell to fit exactly the size of the text it contains.
       	   _table.resizeToFitText(false);

            //adjust the size of the TableNode to include the icons.
            Rectangle2D.Float t_size = _table.getBounds();
            int rows = Math.min(8, _table.getRowCount());
			float height = _table.getCaptionHeight() + rows * _table.getRowHeight();
            _table.getColumns().get(0).setWidth(7);
            if (info.isEnum() || info.isInterface())
            	t_size.width = 30;
            _table.resize(t_size.width + 7, height);
        }

        //check which class members are classes derived from other class(es).
        for (MemberInfo info : membersList)
        {
            if(!info.getParent().isEmpty())
            {
                DiagramNode node = diagram.findNodeById(info.getFullName());
                DiagramNode parent = diagram.findNodeById(info.getParent());

                //add inheritance link
                if(node != null && parent != null)
                {
                    DiagramLink link = diagram.getFactory().createDiagramLink(node, parent);
                    link.setShadowOffsetY(0f);
                    link.setShadowOffsetX(0f);
                }
            }
        }


        buildGroup(enums, "Enums", ":enums");
        buildGroup(interfaces, "Interfaces", ":interfaces");

        arrangeDiagram();
    }

    /**
     *  Build container nodes that hold specific TableNode-s.
     *
     * @param tables   List with the TableNode-s to include in the group.
     * @param name     The name of the group.
     * @param tag      The tag of the group.
     *
     */
    private void buildGroup(ArrayList<TableNode> tables, String name, Object tag)
    {
        if (tables.size() > 0)
        {
            //calculate the location of the node
            int side = (int)Math.ceil(Math.sqrt(tables.size()));
            int rows = (int)Math.ceil((double)tables.size() / side);
            double singleWidth = tables.get(0).getBounds().getWidth();
            double width = 5 * (side + 1) * coef + singleWidth * side;
            double height = (5 * (rows + 1) + 5) * coef;

            double[] rowHeights = new double[rows];
            for (int i = 0; i < tables.size(); i++)
                rowHeights[i / side] = Math.max(tables.get(i).getBounds().getHeight(), rowHeights[i / side]);

            for (double h : rowHeights)
                height += h;


            //initialize the container
            ContainerNode b = diagram.getFactory().createContainerNode(0, 0, width, height);

            int index = 0;
            for (TableNode e : tables)
            {
                b.add(e);

                double x = (index % side) * (5 + singleWidth);
                double y = 0;

                for (int r = 0; r < index / side; r++)
                    y += 5 * coef + rowHeights[r];

                //adjust the size of the node
                e.setBounds(new Rectangle2D.Double(x + 5 * coef, y + 10 * coef,
                        e.getBounds().getWidth(), e.getBounds().getHeight()));

                index++;
            }

            //customize the container
            b.setCaption(name);
            b.setCaptionFormat((new TextFormat(Align.Center, Align.Center)));
            b.setHandlesStyle( HandlesStyle.HatchHandles3 );
            b.setFont(new Font("Verdana", Font.BOLD, 4));
            b.setIgnoreLayout(true);
            b.setTag(tag);
            //no shadow
            b.setShadowOffsetX(0f);
            b.setShadowOffsetY(0f);
            //update the container size.
            b.updateBounds();
        }
    }

    /**
     * Parses info for the classes, enums and interfaces in the selected *.jar.
     *
     * @param pathToJar The url to the jar.
     * @return   A list with MemberInfo objects that describe
     * the API of the provided *.jar file.
     */
    static private ArrayList<MemberInfo> readClasses(String pathToJar)
    {
    	ArrayList<MemberInfo> result = new ArrayList<MemberInfo>();
        try
        {
            //try to open the jar
            JarFile jarFile = new JarFile(pathToJar);

            //explore the elements found in the *.jar.
            Enumeration<JarEntry> e = jarFile.entries();

            URL[] urls = {new URL("jar:file:" + pathToJar + "!/")};
            URLClassLoader cl = URLClassLoader.newInstance(urls);
            //if a class is found - read its data
            while (e.hasMoreElements())
            {
                JarEntry je = e.nextElement();
                if (je.isDirectory() || !je.getName().endsWith(".class")) {
                    continue;
                }
                // -6 because of .class file extension.
                String className = je.getName().substring(0, je.getName().length() - 6);
                className = className.replace('/', '.');

                //omit abstract classes
                if(className.contains("$"))
                    continue;
                Class currClass;
                try
                {
                    //try to create the class object
                	currClass = cl.loadClass(className);
                }
                catch (Error | ClassNotFoundException ex)
                {
                    System.out.print("class could not be created");
                    MemberInfo memberInfo = new MemberInfo(className, className);
                    memberInfo.setPackageName("could not load class");
                    result.add(memberInfo);
                    continue;
                }

                try
            	{
                    //read info only for public classes
                    if(!Modifier.isPublic(currClass.getModifiers()))
                        continue;

                    System.out.println(currClass.getSimpleName());

                    boolean found = false;

                    MemberInfo memberInfo = new MemberInfo(currClass.getSimpleName(), currClass.getName());

                    //raise the respective flags if the class is interface or enum.
                    if(currClass.isInterface())
                        memberInfo.setInterface(true);

                    if(currClass.isEnum())
                    {
                        memberInfo.setEnum(true);
                        //read the members of the enum.
                        List members = Arrays.asList(currClass.getEnumConstants());

                        for(Object obj : members)
                            memberInfo.getFields().add(obj.toString());
                    }

                    //set the name of the package.
                    memberInfo.setPackageName(currClass.getPackage().getName());

                    //to do: check if the super class is within this package
                    if(!memberInfo.isInterface())
                       memberInfo.setParent(currClass.getSuperclass().getName());

                    // get all the constructors of the class
                    Constructor[] constructors = currClass.getDeclaredConstructors();

                    //parse info about the class constructors.
                    for (Constructor ctr : constructors) {

                        String ctrName = ctr.getName();
                        ctrName = ctrName.substring(ctrName.lastIndexOf('.') + 1);
                        ctrName += "(";

                        Parameter[] parameters = ctr.getParameters();

                        int paramCount = 0;

                        //parse the constructor parameters
                        for(Parameter parameter : parameters)
                        {
                            String param = parameter.getType().getSimpleName();
                            int lastDotIdex = param.lastIndexOf('.');

                            if(lastDotIdex > -1)
                                ctrName += param.substring(lastDotIdex);
                            else
                                ctrName += param;

                            paramCount++;

                            if(paramCount < parameters.length)
                                ctrName += ", ";

                        }

                        ctrName += ")";

                        memberInfo.getConstructors().add(ctrName);
                    }


                    //read info about the methods
                    Method[] allMethods = currClass.getDeclaredMethods();
                    for (Method method : allMethods) {
                        //only public methods are considered
                        if (Modifier.isPublic(method.getModifiers())) {

                            String representation = method.getReturnType().getSimpleName() + " " +
                                    method.getName() + "(";

                            //read the parameters
                            Parameter[] parameters = method.getParameters();

                            int paramCount = 0;

                            //parse info about the parameters
                            for(Parameter parameter : parameters)
                            {
                                String param = parameter.getType().getSimpleName();
                                int lastDotIdex = param.lastIndexOf('.');

                               if(lastDotIdex > -1)
                                    representation += param.substring(lastDotIdex);
                                else
                                    representation += param;

                                paramCount++;

                                if(paramCount < parameters.length)
                                    representation += ", ";

                            }

                            representation += ")";

                            Class<?>[] exceptions = method.getExceptionTypes();

                            paramCount = 0;

                            //check if methods throw exception(s)
                            for(Class<?> exception: exceptions)
                            {
                                if(paramCount == 0)
                                    representation += " throws ";
                                paramCount++;

                                representation += exception.getSimpleName();

                                if(paramCount < exceptions.length)
                                    representation += ", ";
                            }

                            //assign the accumulated method name to the MemberInfo object.
                            memberInfo.getMethods().add(representation);
                        }
                    }

                    result.add(memberInfo);
            	}
            	catch (Error err)
            	{
            		
            	}

            }
        }
        catch(java.io.IOException e)
        {
            System.out.println("something went wrong reading the jar");
        }

        return result;
    }

    /**
     * The method arranges the TableNodes using the JDiagram
     * library predefined layout algorithms.
     */
    private void arrangeDiagram()
    {
        //the basic layout is the TreeLayout
        TreeLayout layout = new TreeLayout();

        //the layout type is Centered
        layout.setType(TreeLayoutType.Centered);
        //allow reversed links
        layout.setReversedLinks(true);
        //the type of links will be cascading
        layout.setLinkStyle(TreeLayoutLinkType.Cascading3);
        //specify the distance between levels of tree nodes
        layout.setLevelDistance(25);
        //groups must be preserved.
        layout.setKeepGroupLayout(true);

        layout.arrange(diagram);

        // Place enums and delegates at the end
        DiagramNode enums = diagram.findNode(":enums");
        DiagramNode delegates = diagram.findNode(":interfaces");

        double x = 0;
        //calculate the location of each node to find out the last one
        for (DiagramNode node : diagram.getNodes())
        {
        	if (node instanceof TableNode)
            {
        		x = Math.max(x, node.getBounds().getX() + node.getBounds().getWidth());
            }
        }

        //move enums to the right of the last class node.
        if (enums != null)
        {
            enums.moveTo((float)x + 5f, (float)5);
            x = enums.getBounds().getX() + enums.getBounds().getWidth();
        }

        if (delegates != null)
        {
            delegates.moveTo((float)x + 5, 5);
        }

        diagram.resizeToFitItems(5);
    }

    /**
     * Returns an ImageIcon, or null if the path was invalid.
     *
     * @param path   The path to the image.
     * @param description  The description of the icon.
     * @return  The newly created ImageIcon object.
     */
    protected ImageIcon createImageIcon(String path,
                                        String description) {
        java.net.URL imgURL = getClass().getResource(path);
        if (imgURL != null) {
            return new ImageIcon(imgURL, description);
        } else {
            System.err.println("Couldn't find file: " + path);
            return null;
        }
    }

    /**
     * Runs the application.
     * @param args
     */
    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                MainWindow window = null;
                try {
                    window = new MainWindow();
                    window.setVisible(true);
                }
                catch (Exception exp) {
                    exp.printStackTrace();
                }
            }
        });
    }
}