import java.util.ArrayList;

/**
 * Created by MindFusion on 12/19/2016.
 */
public class MemberInfo {

    private String name;
    private String fullName;
    private String packageName;
    private ArrayList<String> methods;
    private ArrayList<String> fields;
    private ArrayList<String> constructors;
    private String inheritsFrom;
    private boolean isInterface;
    private boolean isEnum;

    /**
     * Initializes a new instance of the MemberInfo class.
     *
     * @param name   The name of the class.
     * @param fullName     The full name including the path.
     */
    public MemberInfo( String name, String fullName )
    {
        this.name = name;
        this.fullName = fullName;

        //lists for the methods, fields and constructors.
        methods = new ArrayList<String>();
        fields = new ArrayList<String>();
        constructors = new ArrayList<String>();
        inheritsFrom = "";
        packageName = "";
        isInterface = false;
        isEnum = false;
    }

    /*
    returns the total count of public members in the class.
     */
    public int memberCount()
    {
        return methods.size() + fields.size() + constructors.size();
    }

    /**
     *
     * @return a list with the methods.
     */
    public ArrayList<String> getMethods()
    {
        return methods;
    }

    /**
     *
     * @return A list with the constructors.
     */
    public ArrayList<String> getConstructors()
    {
        return constructors;
    }

    /**
     *
     * @return A list with the fields.
     */
    public ArrayList<String> getFields()
    {
        return fields;
    }

    /**
     *
     * @return  The full name of the class.
     */
    public String getFullName() {return fullName;}

    /**
     *
     * @param value  true if this is an enum; otherwise false.
     */
    public void setEnum( boolean value)
    {
        this.isEnum = value;
    }

    /**
     *
     * @param value   true if this is an interface; otherwise false.
     */
    public void setInterface( boolean value)
    {
        this.isInterface = value;
    }

    public void setParent(String value)
    {
        this.inheritsFrom = value;
    }

    /**
     *
     * @param value  the name of the package.
     */
    public void setPackageName(String value)
    {
        this.packageName = value;
    }

    /**
     *
     * @return   Gets the name of the package.
     */
    public String getPackageName()
    {
        return packageName;
    }

    /**
     *
     * @return The name of the class.
     */
    public String getName()
    {
        return name;
    }

    /**
     *
     * @return true if this is an enum; otherwise false.
     */
    public boolean isEnum()
    {
        return this.isEnum;
    }

    /**
     *
     * @return   true if this is an interface; otherwise false.
     */
    public boolean isInterface()
    {
        return this.isInterface;
    }

    /**
     *
     * @return The parent of the class.
     */
    public String getParent()
    {
        return inheritsFrom;
    }
}
